#include "plog/init.h"

#include "plog/monoappender_logger.h"
#include "plog/custom_log_proxy_logger.h"
#include "plog/logger_holder.h"

#include "plog/csv_formatter.h"
#include "plog/txt_formatter.h"

#include "plog/debug_output_appender.h"
#include "plog/file_appender.h"
#include "plog/ostream_appender.h"

#include "plog/error_observer.h"
#include "plog/debug_error_observer.h"

#include <memory>
#include <stdexcept>
#include <iostream>

//////////////////////////////////////////////////////////////////////////

namespace 
{
    class csv_signlefile_initializer: public plog::initializer
    {
    public:
        csv_signlefile_initializer(const wchar_t* filename, const plog::log_level level, bool rewrite_file)
            : filename_(filename)
            , level_(level)
            , rewrite_file_(rewrite_file)
        {
        }

        plog::logger* create()
        {
            std::auto_ptr<plog::file_appender> log_appender( new plog::file_appender );
            log_appender->init(filename_, rewrite_file_);
            log_appender->set_level(level_);

            return new plog::monoappender_logger( log_appender.release(), new plog::csv_formatter() );
        }

        void destroy(plog::logger* l)
        {
            if (l)
                delete l;
        }

    private:
        const wchar_t* filename_;
        const plog::log_level level_;
        bool rewrite_file_;
    };

    class txt_signlefile_initializer: public plog::initializer
    {
    public:
        txt_signlefile_initializer(const wchar_t* filename, const plog::log_level level, bool rewrite_file)
            : filename_(filename)
            , level_(level)
            , rewrite_file_(rewrite_file)
        {
        }

        plog::logger* create()
        {
            std::auto_ptr<plog::file_appender> log_appender( new plog::file_appender );
            log_appender->init(filename_, rewrite_file_);
            log_appender->set_level(level_);

            return new plog::monoappender_logger( log_appender.release(), new plog::txt_formatter() );
        }

        void destroy(plog::logger* l)
        {
            if (l)
                delete l;
        }

    private:
        const wchar_t* filename_;
        const plog::log_level level_;
        bool rewrite_file_;
    };

    class ostream_stream_initializer: public plog::initializer
    {
    public:
        ostream_stream_initializer( const plog::log_level level, 
                                    std::ostream& stream, 
                                    plog::formatter* fmtr = NULL)
            : level_(level),
              stream_(stream),
              formatter_(fmtr)
        {
            if (formatter_.get() == NULL)
            {
                formatter_ = std::auto_ptr<plog::formatter>(new plog::txt_formatter);
            }
        }

        plog::logger* create()
        {
            std::auto_ptr<plog::ostream_appender> log_appender( 
                new plog::ostream_appender(stream_, level_) );
            
            return new plog::monoappender_logger( log_appender.release(), formatter_.release() );
        }

        void destroy(plog::logger* l)
        {
            if (l)
                delete l;
        }

    private:
        const plog::log_level           level_;
        std::ostream&                   stream_;
        std::auto_ptr<plog::formatter>  formatter_;
    };

    class debug_output_initializer: public plog::initializer
    {
    public:
        debug_output_initializer(const plog::log_level level)
            : level_(level)
        {
        }

        plog::logger* create()
        {
            std::auto_ptr<plog::debug_output_appender> log_appender( new plog::debug_output_appender(level_) );
            
            return new plog::monoappender_logger( log_appender.release(), new plog::txt_formatter() );
        }

        void destroy(plog::logger* l)
        {
            if (l)
                delete l;
        }

    private:
        const plog::log_level level_;
    };

    class custom_log_initializer: public plog::initializer
    {
    public:
        custom_log_initializer(plog::custom_log* log)
            : log_(log)
        {
        }

        plog::logger* create()
        {
            return new plog::custom_log_proxy_logger(log_);
        }

        void destroy(plog::logger* l)
        {
            if (l)
                delete l;
        }

    private:
        plog::custom_log* log_;
    };
}

//////////////////////////////////////////////////////////////////////////

void plog::init(const wchar_t* filename, const plog::log_level level, bool rewrite_file)
{
    plog::init_txt(filename, level, rewrite_file);
}

void plog::init_csv(const wchar_t* filename, const plog::log_level level, bool rewrite_file)
{
    std::auto_ptr<initializer> li(new csv_signlefile_initializer(filename, level, rewrite_file));
    plog::init(li);
}

void plog::init_txt(const wchar_t* filename, const plog::log_level level, bool rewrite_file)
{
    std::auto_ptr<initializer> li(new txt_signlefile_initializer(filename, level, rewrite_file));
    plog::init(li);
}

void plog::init_ostream(std::ostream& stream, const plog::log_level level, plog::formatter* fmtr)
{
    std::auto_ptr<initializer> li(new ostream_stream_initializer(level, stream, fmtr));
    plog::init(li);
}

void plog::init_cerr(const plog::log_level level)
{
    plog::init_ostream(std::cerr, level);
}

void plog::init_debug(const plog::log_level level)
{
    std::auto_ptr<initializer> li(new debug_output_initializer(level));
    plog::init(li);
}

void plog::init_custom_log(custom_log* log)
{
    std::auto_ptr<initializer> li(new custom_log_initializer(log));
    plog::init(li);
}

std::auto_ptr<plog::initializer> g_li;

void plog::init(std::auto_ptr<initializer> li)
{
    if (plog::get_error_observer() == NULL)
    {
        std::auto_ptr<error_observer> observer(new debug_output_error_observer);
        plog::set_error_observer(observer.release());
    }

    if (plog::logger_holder::get_logger())
    {
        plog::get_error_observer()->on_error("Log has been already initialized");
        throw std::runtime_error("Log has been already initialized");
    }

    g_li = li;
    plog::logger_holder::set_logger(g_li->create());
}

void plog::free(bool free_resources)
{
    if(! free_resources)
    {
        plog::logger_holder::release_logger();
        g_li.release();
        plog::release_error_observer();
    }
    else
    {
        if (g_li.get())
        {
            g_li->destroy(plog::logger_holder::release_logger());
            g_li.reset();
        }
        plog::set_error_observer(NULL);
    }
}
