#define BOOST_TEST_NO_MAIN
#include <boost/test/unit_test.hpp>

#include <boost/filesystem.hpp>

#include "plog/error_observer.h"

#include "plog/init.h"
#include "plog/log.h"

#include <sstream>
#include <iostream>
#include <stdexcept>

namespace
{
    class mock_error_observer
        : public plog::error_observer
    {
    public:
        mock_error_observer()
        {
            ++instances_count_;
        }

        ~mock_error_observer()
        {
            --instances_count_;
        }

        void on_error(const char* error_msg)
        {
            output_ << error_msg;
        }

        std::string get_output_str()
        {
            return output_.str();
        }

        static bool check_instances_count()
        {
            return (instances_count_ == 0);
        }

    private:
        std::stringstream  output_;
        static int         instances_count_;
    };

    int mock_error_observer::instances_count_ = 0;

    struct fixture_error_observer
    {
        fixture_error_observer()
            : observer(new mock_error_observer),
              pobserver(observer.get())
        {}

        ~fixture_error_observer()
        {
            plog::free();
        }

        std::auto_ptr<mock_error_observer> observer;
        mock_error_observer*               pobserver;
    };

    struct fixture_plog
    {
        ~fixture_plog()
        {
            plog::free();
        }
    };

}

BOOST_AUTO_TEST_SUITE(uninitialized_error_observer_test)

BOOST_AUTO_TEST_CASE(uninitialized_error_observer)
{
    BOOST_REQUIRE_EQUAL(plog::get_error_observer(), (plog::error_observer*)NULL);
}

BOOST_AUTO_TEST_SUITE_END()


BOOST_FIXTURE_TEST_SUITE(initialization_error_observer_test, fixture_error_observer)

BOOST_AUTO_TEST_CASE(initialize_default_error_observer)
{
    plog::init_ostream(std::cout, plog::debug);

    BOOST_REQUIRE_NE(plog::get_error_observer(), (plog::error_observer*)NULL);
}

BOOST_AUTO_TEST_CASE(initialize_nondefault_error_observer)
{
    plog::set_error_observer(observer.release());

    BOOST_REQUIRE_EQUAL(plog::get_error_observer(), pobserver);

    plog::init_ostream(std::cout, plog::debug);

    BOOST_REQUIRE_EQUAL(plog::get_error_observer(), pobserver);
}

BOOST_AUTO_TEST_CASE(setting_error_observer_after_log_initialization)
{
    plog::set_error_observer(observer.release());

    plog::init_ostream(std::cout, plog::debug);

    BOOST_REQUIRE_THROW( plog::set_error_observer(pobserver), std::runtime_error );
}

BOOST_AUTO_TEST_CASE(writing_message_to_error_observer)
{
    plog::set_error_observer(observer.release());

    std::stringstream s;
    plog::init_ostream(s, plog::debug);

    // init_ostream() writes error message into error_observer
    BOOST_CHECK_THROW( plog::init_ostream(std::cout, plog::debug), std::runtime_error );

    BOOST_REQUIRE_EQUAL( pobserver->get_output_str(), 
                         std::string("Log has been already initialized") );
}

BOOST_AUTO_TEST_SUITE_END()


BOOST_FIXTURE_TEST_SUITE(destruction_error_observer_test, fixture_plog)

BOOST_AUTO_TEST_CASE(destruction_without_releasing_error_observer)
{
    std::auto_ptr<plog::error_observer> observer(new mock_error_observer);
    
    plog::set_error_observer(observer.release());

    plog::init_ostream(std::cout, plog::debug);

    plog::free( true );

    BOOST_REQUIRE(mock_error_observer::check_instances_count());
}

BOOST_AUTO_TEST_CASE(destruction_with_releasing_error_observer)
{
    plog::error_observer* pobserver = new mock_error_observer;

    plog::set_error_observer(pobserver);

    plog::init_ostream(std::cout, plog::debug);

    plog::release_error_observer();
    BOOST_REQUIRE_EQUAL(plog::get_error_observer(), (plog::error_observer*)NULL);

    delete pobserver;

    BOOST_REQUIRE(mock_error_observer::check_instances_count());
}

BOOST_AUTO_TEST_SUITE_END()
