#define BOOST_TEST_NO_MAIN
#include <boost/test/unit_test.hpp>

#include <memory>

#include "plog/appender.h"
#include "plog/monoappender_logger.h"
#include "plog/csv_formatter.h"
#include "plog/log_message.h"

class MOC_appender 
    : public plog::appender
{
public:
    typedef void (*on_write_callback)(const void*, size_t);

    MOC_appender(plog::log_level min_severity = plog::none, on_write_callback callback = NULL)
        : appender(min_severity)
        , on_write_callback_(callback)
    {
        instances_count_++;
    }

    ~MOC_appender()
    {
        instances_count_--;
    }

    void write(const void* buffer, size_t length)
    {
        if(on_write_callback_)
        {
            on_write_callback_(buffer, length);
        }
    }

    bool newly_created()
    {
        return true;
    }

    static void check_instances_count()
    {
        BOOST_REQUIRE_EQUAL(instances_count_, 0);
    }
private:
    static int  instances_count_;
    on_write_callback  on_write_callback_;
};

int  MOC_appender::instances_count_ = 0;

//////////////////////////////////////////////////////////////////////////
// tests

BOOST_AUTO_TEST_SUITE(non_singleton_log_test)

BOOST_AUTO_TEST_CASE(writing_in_monoappender_logger_without_appenders)
{
    plog::monoappender_logger logger( NULL );
    plog::message message(plog::fatal, CONTEXT_INFO(NULL));

    BOOST_REQUIRE_NO_THROW( logger.write(message) );
}

BOOST_AUTO_TEST_CASE(severity_of_logger_without_appenders)
{
    plog::monoappender_logger logger( NULL );

    BOOST_REQUIRE_EQUAL( logger.check_severity(plog::fatal), false );
}

BOOST_AUTO_TEST_CASE(severity_of_logger_with_appender)
{
    std::auto_ptr<plog::monoappender_logger> logger(
                        new plog::monoappender_logger(
                                        new MOC_appender(plog::warning, NULL) 
                                                     )
                                                    );

    BOOST_REQUIRE_EQUAL(logger->check_severity(plog::error), true);
    BOOST_REQUIRE_EQUAL(logger->check_severity(plog::warning), true);
    BOOST_REQUIRE_EQUAL(logger->check_severity(plog::info), false);

    logger.reset();

    MOC_appender::check_instances_count();
}

namespace
{
    bool g_header_printed = false;

    void check_header(const void* buffer, size_t size)
    {
        plog::csv_formatter formatter;

        BOOST_REQUIRE_EQUAL_COLLECTIONS(  reinterpret_cast<const char*>(buffer)
            , reinterpret_cast<const char*>(buffer) + size
            , formatter.header()
            , formatter.header() + formatter.header_size() );

        BOOST_REQUIRE_EQUAL( size, formatter.header_size() );

        g_header_printed = true;
    }
}

BOOST_AUTO_TEST_CASE(adding_header_to_log)
{
    g_header_printed = false;
    plog::monoappender_logger logger( new MOC_appender(plog::warning, check_header) );

    BOOST_REQUIRE_EQUAL(g_header_printed, true);
    g_header_printed = false;
}

namespace
{
    int g_appended_messages = 0;

    void check_write(const void* buffer, size_t size)
    {
        BOOST_REQUIRE_NE(buffer, (const void*)NULL);

        BOOST_REQUIRE_NE(size, size_t(0));

        g_appended_messages++;
    }
}

BOOST_AUTO_TEST_CASE(writing_to_the_log)
{
    g_appended_messages = 0;

    plog::monoappender_logger logger( new MOC_appender(plog::debug, check_write) );

    BOOST_REQUIRE_EQUAL(g_appended_messages, 1);
    
    plog::message message(plog::info, CONTEXT_INFO(NULL));
    message << "test-test";
    logger.write( message );
    BOOST_REQUIRE_EQUAL(g_appended_messages, 2);

    g_appended_messages = 0;
}

BOOST_AUTO_TEST_CASE(check_for_leaks_of_appender)
{
    MOC_appender::check_instances_count();
}

BOOST_AUTO_TEST_SUITE_END()